implementation module CopyFile

import StdEnv, StdFile, Directory

CopyFile :: String String *env -> *env | FileSystem env
CopyFile inputfname outputfname filesys
| readok && writeok && closeok = finalfilesystem
| not readok  = abort( "cannot open input file: '"+++inputfname+++"'")
| not writeok = abort( "cannot open output file: '"+++outputfname+++"'")
| not closeok = abort( "cannot close output file: '"+++outputfname+++"'")
where
	(readok, inputfile, touchedfilesys ) = sfopen inputfname FReadData filesys
	(writeok, outputfile, nwfilesys) 	  = fopen outputfname FWriteData touchedfilesys
	copiedfile = CharFileCopy inputfile outputfile  
	(closeok, finalfilesystem) 				 = fclose copiedfile nwfilesys


CharListWrite :: [Char] *File -> *File
// writes a list of chars onto a file
CharListWrite [] f = f
CharListWrite [c:cs] f = CharListWrite cs ( fwritec c f )

CharFileCopy :: File *File -> *File
// copies the chars of an opened file to another one
CharFileCopy infile outfile = CharListWrite (CharListRead infile) outfile

CharListRead :: File -> [Char]
// reads a list of chars from a file
CharListRead f
| not readok = []
| otherwise = [char: CharListRead filewithchangedpointer ]
where
	( readok, char, filewithchangedpointer ) = sfreadc f
	
	
createNestedDirectories :: Path *World -> (DirError, *World)
// creates nested directories ( from the current directory, goes through the
// path and creates all the directories mentioned
createNestedDirectories path world 
	= createNestedDirectoriesLevel path 1 world

CanRecoverFrom NoDirError = True
CanRecoverFrom AlreadyExists = True
CanRecoverFrom _ = False

createNestedDirectoriesLevel :: Path Int *World -> (DirError, *World)
createNestedDirectoriesLevel (RelativePath dirnames) i world  
		| i > length( dirnames ) 
			= ( NoDirError, world )
		| i <= length( dirnames ) 
		# ( error, world ) = createDirectory (RelativePath (take i dirnames)) world
		| not( CanRecoverFrom error ) 
			= ( error, world )
		| otherwise 
			= createNestedDirectoriesLevel (RelativePath dirnames) (i+1) world
createNestedDirectoriesLevel (AbsolutePath drivename dirnames) i world  
		| i > length( dirnames ) 
			= ( NoDirError, world )
		| i <= length( dirnames ) 
		# ( error, world ) = createDirectory (AbsolutePath drivename (take i dirnames)) world
		| not( CanRecoverFrom error ) 
			= ( error, world )
		| otherwise 
			= createNestedDirectoriesLevel (AbsolutePath drivename dirnames) (i+1) world


(\+\) infixl 6 :: Path Path -> Path
// concatenates the two paths
(\+\) (AbsolutePath drivename steps) path
	= AbsolutePath drivename (steps ++ PathSteps path)
(\+\) (RelativePath steps) path
	= RelativePath (steps ++ PathSteps path)
	

PathSteps :: Path -> [PathStep]
PathSteps (RelativePath dirnames) = dirnames
PathSteps (AbsolutePath drivename dirnames) = [ PathDown drivename : dirnames ]

CutLast :: Path -> (Path, Path)
CutLast path
	= ( ButLast path, RelativePath [last (PathSteps path)] )

		 
IsError :: DirError -> Bool
IsError NoDirError = False
IsError _ = True 


ErrorType :: DirError -> String
ErrorType NoDirError = " no dir error"
ErrorType DoesntExist = " doesnt exist"
ErrorType BadName = " bad name"
ErrorType NotEnoughSpace = " not enough space"
ErrorType AlreadyExists = " already exists"
ErrorType NoPermission = " no permission"
ErrorType MoveIntoOffspring = " move into offspring"
ErrorType MoveAcrossDisks = " move across disks"
ErrorType NotYetRemovable = " not yet removable"
ErrorType OtherDirError = " other dir error"

ButLast :: Path -> Path
ButLast (RelativePath steps) = RelativePath ( init steps )
ButLast (AbsolutePath drivename steps) = AbsolutePath drivename (init steps )

FileNameOf :: Path -> String
FileNameOf path =
	case (last (PathSteps path)) of
		 PathDown name -> name
		 PathUp -> ".."

WithoutExtension :: String -> String
WithoutExtension st 
	= st % (0, (DotIndex st)-1)

DotIndex st = DotIndexFrom ((size st)-1) st
DotIndexFrom i st 
	| i < 0 = size st 
	| st.[i] == '.' = i
	| st.[i] == '/' || st.[i] == '\\' = size st
	| otherwise = DotIndexFrom (i-1) st

HasExtension :: String -> Bool
HasExtension st = not ( (DotIndex st) == (size st) )

		